/*-
 * Copyright (c) 2002-2004 Norikatsu Shigemura
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: linux_file.c,v 1.5 2004/02/28 12:52:40 nork Exp $
 */

/*
 *
 */
#include "pluginwrapper.h"
#include "linux.h"
#include <errno.h>
#include <stdarg.h>
#include <fcntl.h>
#include <stddef.h>


/*
 * open(2) system call
 */
extern int _open(const char *, int, ...);

int
open(const char *path, int l_flags, .../* mode_t mode */)
{
	va_list		 ap;
	int		 ret, flags = 0;
	const char	*name = NULL;

	va_start(ap, l_flags);

	CLEAR_ERRNO;
#if 0
	if(  strncmp(path, DSP_DEVICE, sizeof(DSP_DEVICE) - 1) == 0  )  {
		if(  ( name = getenv(ENV_DSPDEV) )  !=  NULL  )  {
			struct stat st;
			if(  stat(name, &st) == 0  )  {
				if(  ( S_ISCHR(st.st_mode) || S_ISBLK(st.st_mode) )
				  && major(st.st_rdev) == DEV_SND_MAJOR  )  {
					path = name;
				}
			}  else  { errno = 0; }
		}
	}
#endif

	if(  l_flags & LINUX_O_RDONLY    )  flags |= O_RDONLY;
	if(  l_flags & LINUX_O_WRONLY    )  flags |= O_WRONLY;
	if(  l_flags & LINUX_O_RDWR      )  flags |= O_RDWR;
	if(  l_flags & LINUX_O_NDELAY    )  flags |= O_NONBLOCK;
	if(  l_flags & LINUX_O_APPEND    )  flags |= O_APPEND;
	if(  l_flags & LINUX_O_SYNC      )  flags |= O_FSYNC;
	if(  l_flags & LINUX_O_NONBLOCK  )  flags |= O_NONBLOCK;
	if(  l_flags & LINUX_FASYNC      )  flags |= O_ASYNC;
	if(  l_flags & LINUX_O_CREAT     )  flags |= O_CREAT;
	if(  l_flags & LINUX_O_TRUNC     )  flags |= O_TRUNC;
	if(  l_flags & LINUX_O_EXCL      )  flags |= O_EXCL;
	if(  l_flags & LINUX_O_NOCTTY    )  flags |= O_NOCTTY;

	if(  flags & O_CREAT  )  {
		ret = _open(path, flags, va_arg(ap, mode_t *));
	}  else  {
		ret = _open(path, flags);
	}

	dprintf("open(path=%s, flags=0%06o(0%06o on linux)) = %d / ERRNO = %d", path, flags, l_flags, ret, errno);

	va_end(ap);

	return ret;
}


/*
 * close(2) system call for DEBUG use
 */
#ifdef DEBUG
extern int _close(int);

int
close(int fd)
{
	int ret;

	CLEAR_ERRNO;

	ret = _close(fd);

	dprintf("close(fd=%d) = %d / ERRNO = %d", fd, ret, errno);

	return ret;
}
#endif


/*
 * read(2) system call for DEBUG use
 */
#ifdef DEBUG
extern ssize_t _read(int, void *, size_t);

ssize_t
read(int fd, void *buf, size_t nbytes)
{
	/* sleep(3600); */
	ssize_t ret;

	CLEAR_ERRNO;

	ret = _read(fd, buf, nbytes);

	dprintf("read(fd=%d, buf=0x%08x, nbytes=%u) = %d / ERRNO = %d", fd, buf, nbytes, ret, errno);

	return ret;
}
#endif


/*
 * write(2) system call for DEBUG use
 */
#ifdef DEBUG
extern ssize_t _write(int, const void *, size_t);

ssize_t
write(int fd, const void *buf, size_t nbytes)
{
	ssize_t ret;

	CLEAR_ERRNO;

	ret = _write(fd, buf, nbytes);

	dprintf("write(fd=%d, buf=0x%08x, nbytes=%u) = %d / ERRNO = %d", fd, buf, nbytes, ret, errno);

	return ret;
}
#endif


/*
 * fcntl(2) system call
 */
int _fcntl(int, int, ...);

int
fcntl(int fd, int l_cmd, ...)
{
	va_list		ap;
	int		arg = 0, l_arg, ret, l_ret = 0;

	va_start(ap, l_cmd);

	CLEAR_ERRNO;

	switch(  l_cmd  )  {
	case LINUX_F_GETFD:
		ret = _fcntl(fd, F_GETFD);
		dprintf("fcntl(fd=%d, LINUX_F_GETFD) = %d / ERRNO = %d", fd, ret, errno);
		break;
		
        case LINUX_F_SETFD:
		ret = _fcntl(fd, F_SETFD, (l_arg = va_arg(ap, int)));
		dprintf("fcntl(fd=%d, LINUX_F_SETFD, arg=%d) = %d / ERRNO =%d", fd, l_arg, ret, errno);
		break;

	case LINUX_F_GETFL:
		ret = _fcntl(fd, F_GETFL);

		if(  ret >= 0  )  {
			if(  ret & O_RDONLY  )  l_ret |= LINUX_O_RDONLY;
			if(  ret & O_WRONLY  )  l_ret |= LINUX_O_WRONLY;
			if(  ret & O_RDWR    )  l_ret |= LINUX_O_RDWR;
			if(  ret & O_NDELAY  )  l_ret |= LINUX_O_NONBLOCK;
			if(  ret & O_APPEND  )  l_ret |= LINUX_O_APPEND;
			if(  ret & O_FSYNC   )  l_ret |= LINUX_O_SYNC;
			if(  ret & O_ASYNC   )  l_ret |= LINUX_FASYNC;
		}  else  {
			l_ret = -1;
		}

		dprintf("fcntl(fd=%d, LINUX_F_GETFL) = 0%06o(0%06o on BSD) / ERRNO = %d", fd, l_ret, ret, errno);

		ret = l_ret;
		break;

        case LINUX_F_SETFL:
		l_arg = va_arg(ap, int);

		if(  l_arg & LINUX_O_NDELAY  )  arg |= O_NONBLOCK;
                if(  l_arg & LINUX_O_APPEND  )  arg |= O_APPEND;
                if(  l_arg & LINUX_O_SYNC    )  arg |= O_FSYNC;
                if(  l_arg & LINUX_FASYNC    )  arg |= O_ASYNC;

		ret = _fcntl(fd, F_SETFL, arg);

		dprintf("fcntl(fd=%d, LINUX_F_SETFL, arg=0%06o(0%06o on BSD)) = %d / ERRNO = %d", fd, l_arg, arg, ret, errno);

		break;

	default:
		errno = ENOSYS;
		ret = -1;
		dprintf("fcntl(fd=%d, cmd=%d, ...) = %d / ERRNO = %d", fd, l_cmd, ret, errno);
		break;
	}

	return ret;
}


/*
 * pipe(2) system call for DEBUG use
 */
#ifdef DEBUG
int _pipe(int *);

int
pipe(int *fildes)
{
	int ret;

	CLEAR_ERRNO;

	ret = _pipe(fildes);

	dprintf("pipe(ret_fd=%d, ret_fd=%d) = %d / ERRNO = %d", fildes[0], fildes[1], ret, errno);

	return ret;
}
#endif


/*
 * dup2(2) system call for DEBUG use
 */
#ifdef DEBUG
int _dup2(int, int);

int
dup2(int oldd, int newd)
{
	int ret;

	CLEAR_ERRNO;

	ret = _dup(oldd, newd);

	dprintf("dup2(%d, %d) = %d / ERRNO = %d", oldd, newd, ret, errno);

	return ret;
}
#endif


/*
 * access(2) system call for DEBUG use
 */
#ifdef DEBUG
int _access(const char *, int);

int
access(const char *path, int mode)
{
	int ret;

	CLEAR_ERRNO;

	ret = _access(path, mode);

	dprintf("access(%s, %o) = %d / ERRNO = %d", path, mode, ret, errno);

	return ret;
}
#endif
