/*-
 * Copyright (c) 2002-2005 Norikatsu Shigemura
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: linux_socket.c,v 1.11 2005/01/19 14:14:05 nork Exp $
 */

/*
 *
 */
#include "pluginwrapper.h"
#include "linux.h"
#include <errno.h>
#include <string.h>
#include <limits.h>
#include <sys/types.h>
#define	bind		dummy_bind
#define	socket		dummy_socket
#define	connect		dummy_connect
#define	listen		dummy_listen
#define	accept		dummy_accept
#define	setsockopt	dummy_setsockopt
#define	getsockopt	dummy_getsockopt
#include <sys/socket.h>
#undef	bind
#undef	socket
#undef	connect
#undef	listen
#undef	accept
#undef	setsockopt
#undef	getsockopt

#include <netinet/in.h>


/*
 *
 */
static int
linux_to_bsd_domain(int l_domain)
{
	switch(  l_domain  )  {
	case LINUX_AF_UNIX:
		return AF_LOCAL;

	default:
		dprintf("linux_to_bsd_domain(%d) = UNKNOWN(-1)", l_domain);
        }

	return -1;
}


/*
 *
 */
static int
linux_to_bsd_sockopt_level(int l_level)
{
	switch(  l_level  )  {
	case LINUX_SOL_SOCKET:
		return SOL_SOCKET;

	default:
		dprintf("linux_to_bsd_sockopt_level(%d) = keep it(%d)", l_level, l_level);
	}

	return l_level;
}


/*
 *
 */
static int
linux_to_bsd_so_sockopt(int l_optname)
{
	switch(  l_optname  )  {
	case LINUX_SO_SNDBUF:
		return SO_SNDBUF;

	case LINUX_SO_RCVBUF:
		return SO_RCVBUF;

	default:
		dprintf("linux_to_bsd_so_sockopt(%d) = UNKNOWN(-1)", l_optname);
	}

	return -1;
}


/*
 *
 */
static int
linux_to_bsd_ip_sockopt(int l_optname)
{

	switch(  l_optname  ) {
	case LINUX_IP_OPTIONS:
		return IP_OPTIONS;

	default:
		dprintf("linux_to_bsd_ip_sockopt(%d) = UNKNOWN(-1)", l_optname);
	}

	return -1;
}


/*
 * bind(2) system call
 */
int _bind(int, const struct sockaddr *, socklen_t);

int
bind(int s, const struct linux_sockaddr *l_addr, int addrlen)
{
	errno = ENOSYS;
	dprintf("bind: not implimented / ERRNO = %d", errno);
	return -1;
}


/*
 * socket(2) system call
 */
int _socket(int, int, int);

int
socket(int l_domain, int type, int protocol)
{
	int domain, ret;

	CLEAR_ERRNO;

	domain = linux_to_bsd_domain(l_domain);
	ret    = _socket(domain, type, protocol);

	dprintf("socket(domain=%d(%d on linux), type=%d, protocol=%d) = %d / ERRNO=%d", domain, l_domain, type, protocol, ret, errno);

	return ret;
}


/*
 * connect(2) system call
 */
int _connect(int, const struct sockaddr *, socklen_t);

int
connect(int s, const struct linux_sockaddr *l_name, int namelen)
{
	int ret, domain;
	struct sockaddr_storage name;

	CLEAR_ERRNO;

	if(  namelen < 2 || namelen > UCHAR_MAX || l_name == NULL  )  {
		errno = EINVAL;
		return -1;
	}

	memcpy(&name, l_name, namelen);

	if(  ( domain = linux_to_bsd_domain(l_name->sa_family) ) == -1  )  {
		errno = EINVAL;
		return -1;
	}  else  {
		name.ss_family = domain;
	}

	if(  name.ss_family == AF_INET  )  {
		name.ss_len = sizeof(struct sockaddr_in);
	}  else  {
		name.ss_len = namelen;
	}

	ret = _connect(s, (const struct sockaddr*)&name, name.ss_len);

#ifdef	DEBUG
	if(  name.ss_family == AF_LOCAL  )  {
		char buf[UCHAR_MAX];
		strlcpy(buf, name.__ss_pad1, namelen - 2 + 1);
		dprintf("connect(s=%d, name=%s, namelen=%d(%d on linux) = %d / ERRNO = %d", s, buf, name.ss_len, namelen, ret, errno);
	}  else  {
		dprintf("connect(s=%d, name=0x%08x(0x%08x on linux), namelen=%d(%d on linux) = %d / ERRNO = %d", s, &name, l_name, name.ss_len, namelen, ret, errno);
	}
#endif

	return ret;
}


/*
 * listen(2) system call for DEBUG use
 */
#ifdef DEBUG
int _listen(int, int);

int
listen(int s, int backlog)
{
	int ret;

	CLEAR_ERRNO;

	ret = _listen(s, backlog);

	dprintf("listen(s=%d, backlog=%d) = %d / ERRNO = %d", s, backlog, ret, errno);

	return ret;
}
#endif


/*
 * accept(2) system call for kludge
 */
int _accept(int, struct sockaddr *, socklen_t *);

int
accept(int s, struct linux_sockaddr *l_addr, int *addrlen)
{
	int ret, domain;

	CLEAR_ERRNO;

	/* XXX: not implemented conversion from sockaddr to osockaddr. */
	if(  s >=  0  )  {
		ret = _accept(s, (struct sockaddr *)l_addr, (socklen_t *)addrlen);
	}  else  {
		ret = -1;
		errno = EBADF;
	}

	dprintf("accept(s=%d, addr=0x%08x, addrlen=%08x) = %d / ERRNO = %d", s, l_addr, addrlen, ret, errno);

	return ret;
}


/*
 * setsockopt(2) system call
 */
int _setsockopt(int, int, int, const void *, socklen_t);

int
setsockopt(int s, int l_level, int l_optname, const void *optval, socklen_t optlen)
{
	int ret, optname, level = linux_to_bsd_sockopt_level(l_level);

	CLEAR_ERRNO;

	switch(  level  ) {
	case SOL_SOCKET:
		optname = linux_to_bsd_so_sockopt(l_optname);
		break;

	case IPPROTO_IP:
		optname = linux_to_bsd_ip_sockopt(l_optname);
		break;

	case IPPROTO_TCP:
		/* Linux TCP option values match BSD's */
		optname = optname;
		break;

	default:
		optname = -1;
		break;
	}

	if(  optname == -1  )  {
		errno = EINVAL;
		ret   = -1;
	}  else  {
		ret = _setsockopt(s, level, optname, optval, optlen);
	}

	dprintf("setsocketopt(s=%d, level=%d(%d on linux), optname=%d(%d on linux), optval=%d, optlen=%d) = %d / ERRNO = %d", s, level, l_level, optname, l_optname, *(int *)optval, optlen, ret, errno);

	return ret;
}


/*
 * getsockopt(2) system call
 */
int _getsockopt(int, int, int, void *, socklen_t *);

int
getsockopt(int s, int l_level, int l_optname, void *optval, socklen_t *optlen)
{
	int ret, optname, level = linux_to_bsd_sockopt_level(l_level);

	CLEAR_ERRNO;

	switch(  level  ) {
	case SOL_SOCKET:
		optname = linux_to_bsd_so_sockopt(l_optname);
		break;

	case IPPROTO_IP:
		optname = linux_to_bsd_ip_sockopt(l_optname);
		break;

	case IPPROTO_TCP:
		/* Linux TCP option values match BSD's */
		optname = l_optname;
		break;

	default:
		optname = -1;
		break;
	}

	if(  optname == -1  )  {
		errno = EINVAL;
		ret   = -1;
	}  else  {
		ret = _getsockopt(s, level, optname, optval, optlen);
	}

	dprintf("getsocketopt(s=%d, level=%d(%d on linux), optname=%d(%d on linux), optval=%0d, optlen=%d(result)) = %d / ERRNO = %d", s, level, l_level, optname, l_optname, *(int*)optval, *optlen, ret, errno);

	return ret;
}
